@@ -1,29 +1,22 @@
 # -*- coding: utf-8 -*-
 # See LICENSE.txt for licensing terms
-#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.12/rst2pdf/flowables.py $
-#$LastChangedDate: 2009-10-19 10:24:25 -0300 (Mon, 19 Oct 2009) $
-#$LastChangedRevision: 1222 $
+#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.14/rst2pdf/flowables.py $
+#$LastChangedDate: 2010-03-25 23:12:19 -0300 (Thu, 25 Mar 2010) $
+#$LastChangedRevision: 2052 $
 
 __docformat__ = 'reStructuredText'
 
+from copy import copy
+
 from reportlab.platypus import *
 from reportlab.platypus.doctemplate import *
 from reportlab.lib.enums import *
 
-try:
-    import wordaxe
-    from wordaxe.rl.paragraph import Paragraph
-    HAS_WORDAXE=True
-except ImportError:
-    from reportlab.platypus.paragraph import Paragraph
-    HAS_WORDAXE=False
+from opt_imports import Paragraph, NullDraw
 
 from reportlab.lib.units import *
-try:
-    from reportlab.platypus.flowables import _listWrapOn, _FUZZ, NullDraw
-except ImportError: # Probably RL 2.1
-    from reportlab.platypus.flowables import _listWrapOn, _FUZZ
-    NullDraw = Flowable
+from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT
+from reportlab.platypus.flowables import _listWrapOn, _FUZZ
 from reportlab.platypus.tableofcontents import TableOfContents
 from reportlab.lib.styles import ParagraphStyle
 
@@ -33,68 +26,6 @@ from log import log
 import re
 from xml.sax.saxutils import unescape, escape
 
-
-class MyImage(Image):
-    """A Image subclass that can take a 'percentage_of_container' kind,
-    which resizes it on wrap() to use... well, a percentage of the
-    container's width.
-
-    This whole class is a huge ungainly hack that deserves flaming death.
-    However, I have asked in the reportlab list for this feature. If they
-    implement it ... hey, I kill this in a jiffie.
-
-    """
-
-    def __init__(self, filename, width=None, height=None,
-                 kind='direct', mask="auto", lazy=1):
-        self.__kind=kind
-        if kind == 'percentage_of_container':
-            Image.__init__(self, filename, width, height,
-                'direct', mask, lazy)
-            self.drawWidth=width
-            self.drawHeight=height
-            self.__width=width
-            self.__height=height
-        else:
-            Image.__init__(self, filename, width, height,
-                kind, mask, lazy)
-        self.__ratio=float(self.imageWidth)/self.imageHeight
-        self.__wrappedonce=False
-
-    def wrap(self, availWidth, availHeight):
-        if self.__kind=='percentage_of_container':
-            w, h= self.__width, self.__height
-            if not w:
-                log.warning('Scaling image as % of container with w unset.'
-                'This should not happen, setting to 100')
-                w = 100
-            scale=w/100.
-            w = availWidth*scale
-            h = w/self.__ratio
-            self.drawWidth, self.drawHeight = w, h
-            return w, h
-        else:
-            if self.drawHeight > availHeight:
-                if not self._atTop:
-                    return Image.wrap(self, availWidth, availHeight)
-                else:
-                    # It's the first thing in the frame, probably
-                    # Wrapping it will not make it work, so we
-                    # adjust by height
-                    # FIXME get rst file info (line number)
-                    # here for better error message
-                    log.warning('image %s is too tall for the '\
-                                'frame, rescaling'%\
-                                self.filename)
-                    self.drawHeight = availHeight
-                    self.drawWidth = availHeight*self.__ratio
-            elif self.drawWidth > availWidth:
-                log.warning('image %s is too wide for the frame, rescaling'%\
-                            self.filename)
-                self.drawWidth = availWidth
-                self.drawHeight = availWidth / self.__ratio
-            return Image.wrap(self, availWidth, availHeight)
-
 class MyIndenter(Indenter):
     """An indenter that has a width, because otherwise you get crashes
     if added inside tables"""
@@ -191,6 +122,36 @@ class Reference(Flowable):
     def __str__(self):
         return "Reference: %s" % self.refid
 
+class OddEven(Flowable):
+    """This flowable takes two lists of flowables as arguments, odd and even.
+    If will draw the "odd" list when drawn in odd pages and the "even" list on
+    even pages.
+
+
+    wrap() will always return a size large enough for both lists, and this flowable
+    **cannot** be split, so use with care.
+    """
+    
+    def __init__(self, odd, even, style=None):
+        self.odd=DelayedTable([[odd]],['100%'], style)
+        self.even=DelayedTable([[even]],['100%'], style)
+
+    def wrap(self, w, h):
+        """Return a box large enough for both odd and even"""
+        w1,h1=self.odd.wrap(w,h)
+        w2,h2=self.even.wrap(w,h)
+        return max(w1,w2), max (h1,h2)
+
+    def drawOn(self, canvas, x, y, _sW=0):
+        if canvas._pagenum %2 == 0:
+            self.even.drawOn(canvas, x, y, _sW)
+        else:
+            self.odd.drawOn(canvas, x, y, _sW)
+
+    def split(self):
+        """Makes no sense to split this..."""
+        return []
+
 class DelayedTable(Flowable):
     """A flowable that inserts a table for which it has the data.
 
@@ -199,12 +160,14 @@ class DelayedTable(Flowable):
 
     """
 
-    def __init__(self, data, colWidths, style, repeatrows=False):
+    def __init__(self, data, colWidths, style, repeatrows=False, splitByRow=True):
         self.data = data
         self._colWidths = colWidths
         self.style = style
         self.t = None
         self.repeatrows = repeatrows
+        self.hAlign = TA_CENTER
+        self.splitByRow=splitByRow
 
     def wrap(self, w, h):
         # Create the table, with the widths from colWidths reinterpreted
@@ -218,14 +181,28 @@ class DelayedTable(Flowable):
         self.colWidths=map(adjust, self._colWidths)
         #colWidths = [_w * _tw for _w in self.colWidths]
         self.t = Table(self.data, colWidths=self.colWidths,
-            style=self.style, repeatRows=self.repeatrows)
+            style=self.style, repeatRows=self.repeatrows, 
+            splitByRow=True)
+            #splitByRow=self.splitByRow)
+        self.t.hAlign = self.hAlign
         return self.t.wrap(w, h)
 
     def split(self, w, h):
-        return self.t.split(w, h)
+        if self.splitByRow:
+            return self.t.split(w, h)
+        else:
+            return []
 
     def drawOn(self, canvas, x, y, _sW=0):
         self.t.drawOn(canvas, x, y, _sW)
+        
+    def identity(self, maxLen=None):
+        print self.data
+        return "<%s at %s%s%s> containing: %s" % (self.__class__.__name__,
+            hex(id(self)), self._frameName(),
+            getattr(self, 'name', '')
+                and (' name="%s"' % getattr(self, 'name', '')) or '',
+                unicode(self.data[0])[:180])
 
 def tablepadding(padding):
     if not isinstance(padding,(list,tuple)):
@@ -242,10 +219,10 @@ class SplitTable(DelayedTable):
             sys.exit(1)
         DelayedTable.__init__(self,data,colWidths,style)
         self.padding, p1, p2, p3, p4=tablepadding(padding)
-        self.style.add(*p1)
-        self.style.add(*p2)
-        self.style.add(*p3)
-        self.style.add(*p4)
+        self.style._cmds.insert(0,p1)
+        self.style._cmds.insert(0,p2)
+        self.style._cmds.insert(0,p3)
+        self.style._cmds.insert(0,p4)
 
     def identity(self, maxLen=None):
         return "<%s at %s%s%s> containing: %s" % (self.__class__.__name__,
@@ -267,9 +244,6 @@ class SplitTable(DelayedTable):
             bullet=self.data[0][0]
             text=self.data[0][1]
             for l in range(0,len(text)):
-                # The -12 here is a hack. I don't know why
-                # If the table contents wrap to Y points, the
-                # table wraps to Y+12
                 _,fh = _listWrapOn(text[:l+1],w-dw,None)
                 if fh+dh > h:
                     # The lth flowable is the guilty one
@@ -279,18 +253,39 @@ class SplitTable(DelayedTable):
                     # Workaround for Issue 180
                     text[l].wrap(w-dw,h-lh-dh)
                     l2=text[l].split(w-dw,h-lh-dh)
-                    if l2==[] and l==0: # Can't fit anything
-                        return l2
-                    elif l2==[]: # Not splittable, push to next page
+                    if l2==[]: # Not splittable, push some to next page
+                        if l==0: # Can't fit anything, push all to next page
+                            return l2
+ 
+                        # We reduce the number of items we keep on the
+                        # page for two reasons:
+                        #    1) If an item is associated with the following
+                        #       item (getKeepWithNext() == True), we have
+                        #       to back up to a previous one.
+                        #    2) If we miscalculated the size required on
+                        #       the first page (I dunno why, probably not
+                        #       counting borders properly, but we do
+                        #       miscalculate occasionally).  Seems to
+                        #       have to do with nested tables, so it might
+                        #       be the extra space on the border on the
+                        #       inner table.
 
-                        # If the previous one is keepwithnext, push
-                        # that one too
-                        while l>0 and text[l-1].getKeepWithNext():
-                            l-=1
+                        while l > 0:
+                            if not text[l-1].getKeepWithNext():
+                                first_t = Table([
+                                                [bullet,
+                                                text[:l]]
+                                            ],
+                                            colWidths=self.colWidths,
+                                            style=self.style)
+                                _w,_h = first_t.wrap(w, h)
+                                if _h <= h:
+                                    break
+                            l -= 1
 
                         if l>0:
                             # Workaround for Issue 180 with wordaxe:
-                            #if HAS_WORDAXE:
+                            #if wordaxe is not None:
                                 #l3=[Table([
                                             #[bullet,
                                              #text[:l]]
@@ -301,12 +296,7 @@ class SplitTable(DelayedTable):
                                         #colWidths=self.colWidths,
                                         #style=self.style)]
                             #else:
-                            l3=[Table([
-                                        [bullet,
-                                         text[:l]]
-                                       ],
-                                    colWidths=self.colWidths,
-                                    style=self.style),
+                            l3=[first_t,
                                     SplitTable([['',text[l:]]],
                                     colWidths=self.colWidths,
                                     style=self.style,
@@ -320,7 +310,7 @@ class SplitTable(DelayedTable):
                                 style=self.style)]
                         if l2[1:]+text[l+1:]:
                             ## Workaround for Issue 180 with wordaxe:
-                            #if HAS_WORDAXE:
+                            #if wordaxe is not None:
                                 #l3.append(
                                     #Table([['',l2[1:]+text[l+1:]]],
                                     #colWidths=self.colWidths,
@@ -365,7 +355,6 @@ class MyPageBreak(FrameActionFlowable):
 
     def frameAction(self, frame):
         frame._generated_content = []
-
         if self.breakTo=='any': # Break only once. None if at top of page
             if not frame._atTop:
                 frame._generated_content.append(SetNextTemplate(self.templateName))
@@ -507,15 +496,15 @@ class SmartFrame(Frame):
             id, showBoundary, overlapAttachedSpace, _debug)
 
     def add (self, flowable, canv, trySplit=0):
-        if self._atTop:
-            flowable._atTop=True
-        else:
-            flowable._atTop=False
+        flowable._atTop=self._atTop
         return Frame.add(self, flowable, canv, trySplit)
 
     def __repr__(self):
         return self.__s
 
+    def __deepcopy__(self, *whatever):
+        return copy(self)
+
 class FrameCutter(FrameActionFlowable):
 
     def __init__(self, dx, width, flowable, padding, lpad, floatLeft=True):
@@ -652,6 +641,16 @@ class BoundByWidth(Flowable):
         self.pad = None
         Flowable.__init__(self)
 
+    def border_padding(self, useWidth, additional):
+        sdict = self.style
+        sdict = sdict.__dict__ or {}
+        bp = sdict.get("borderPadding", 0)
+        if useWidth:
+            additional += sdict.get("borderWidth", 0)
+        if not isinstance(bp, list):
+            bp = [bp] * 4
+        return [x + additional for x in bp]
+
     def identity(self, maxLen=None):
         return "<%s at %s%s%s> containing: %s" % (self.__class__.__name__,
             hex(id(self)), self._frameName(),
@@ -661,21 +660,7 @@ class BoundByWidth(Flowable):
 
     def wrap(self, availWidth, availHeight):
         """If we need more width than we have, complain, keep a scale"""
-        if self.style:
-            bp = self.style.__dict__.get("borderPadding", 0)
-            bw = self.style.__dict__.get("borderWidth", 0)
-            if isinstance(bp,list):
-                self.pad = [bp[0] + bw + .1,
-                            bp[1] + bw + .1,
-                            bp[2] + bw + .1,
-                            bp[3] + bw + .1]
-            else:
-                self.pad = [bp + bw + .1,
-                            bp + bw + .1,
-                            bp + bw + .1,
-                            bp + bw + .1]
-        else:
-            self.pad = [0,0,0,0]
+        self.pad = self.border_padding(True, 0.1)
         maxWidth = float(min(
             styles.adjustUnits(self.maxWidth, availWidth) or availWidth,
                                availWidth))
@@ -786,13 +771,13 @@ class BoxedContainer(BoundByWidth):
             fill=1
         else:
             fill=0
-        if self.style:
-            self.padding = self.style.__dict__.get('padding', 8)
-        else:
-            self.padding = 0
-        self.padding += lw
+
+
+        padding = self.border_padding(False, lw)
+        xpadding = padding[1] + padding[3]
+        ypadding = padding[0] + padding[2]
         p = canv.beginPath()
-        p.rect(x, y, self.width + 2*self.padding, self.height + 2*self.padding)
+        p.rect(x, y, self.width + xpadding, self.height + ypadding)
         canv.drawPath(p, stroke=stroke, fill=fill)
         canv.restoreState()
         BoundByWidth.draw(self)
