@@ -25,6 +25,7 @@ import os
 
 from cStringIO import StringIO
 from rst2pdf import createpdf
+
 from rst2pdf import pygments_code_block_directive
 from pygments.lexers import get_lexer_by_name, guess_lexer
 
@@ -78,6 +79,14 @@ class PDFBuilder(Builder):
                 else:
                     opts={}
                 self.info("processing " + targetname + "... ", nonl=1)
+                
+                class dummy:
+                    extensions=self.config.pdf_extensions
+                    
+                createpdf.add_extensions(dummy())
+                
+                self.page_template=opts.get('pdf_page_template',self.config.pdf_page_template)
+                
                 docwriter = PDFWriter(self,
                                 stylesheets=opts.get('pdf_stylesheets',self.config.pdf_stylesheets),
                                 language=opts.get('pdf_language',self.config.pdf_language),
@@ -88,12 +97,15 @@ class PDFBuilder(Builder):
                                 compressed=opts.get('pdf_compressed',self.config.pdf_compressed),
                                 inline_footnotes=opts.get('pdf_inline_footnotes',self.config.pdf_inline_footnotes),
                                 splittables=opts.get('pdf_splittables',self.config.pdf_splittables),
-                                default_dpi=opts.get('pdf_default_dpi', self.config.pdf_default_dpi),
+                                default_dpi=opts.get('pdf_default_dpi',self.config.pdf_default_dpi),
+                                page_template=self.page_template,
+                                invariant=opts.get('pdf_invariant',self.config.pdf_invariant),
                                 srcdir=self.srcdir,
                                 config=self.config
                                 )
+                
                 tgt_file = path.join(self.outdir, targetname + self.out_suffix)
-                destination = FileOutput(destination_path=tgt_file, encoding='utf-8')
+                destination = FileOutput(destination=open(tgt_file,'wb'), encoding='utf-8')
                 doctree = self.assemble_doctree(docname,title,author, 
                     appendices=opts.get('pdf_appendices', self.config.pdf_appendices) or [])
                 doctree.settings.author=author
@@ -181,7 +193,7 @@ class PDFBuilder(Builder):
             
             if genindex: # No point in creating empty indexes
                 index_nodes=genindex_nodes(genindex)
-                tree.append(nodes.raw(text='OddPageBreak twoColumn'))
+                tree.append(nodes.raw(text='OddPageBreak twoColumn', format='pdf'))
                 tree.append(index_nodes)
 
         # This is stolen from the HTML builder
@@ -276,11 +288,11 @@ class PDFBuilder(Builder):
                 output.append('')
                 
             dt = docutils.core.publish_doctree('\n'.join(output))[1:]
-            dt.insert(0,nodes.raw(text='OddPageBreak twoColumn'))
+            dt.insert(0,nodes.raw(text='OddPageBreak twoColumn', format='pdf'))
             tree.extend(dt)
                     
         if appendices:
-            tree.append(nodes.raw(text='OddPageBreak cutePage'))
+            tree.append(nodes.raw(text='OddPageBreak %s'%self.page_template, format='pdf'))
             self.info()
             self.info('adding appendixes...', nonl=1)
             for docname in appendices:
@@ -466,6 +478,8 @@ class PDFWriter(writers.Writer):
                 splittables = False,
                 srcdir = '.',
                 default_dpi = 300,
+                page_template = 'cutePage',
+                invariant = 'False',
                 config = {}):
         writers.Writer.__init__(self)
         self.builder = builder
@@ -483,6 +497,8 @@ class PDFWriter(writers.Writer):
         self.srcdir = srcdir
         self.config = config
         self.default_dpi = default_dpi
+        self.page_template = page_template
+        self.invariant=invariant
 
     supported = ('pdf')
     config_section = 'pdf writer'
@@ -505,10 +521,10 @@ class PDFWriter(writers.Writer):
         pending=nodes.topic()
         contents.append(pending)
         pending.details={}
-        self.document.insert(0,nodes.raw(text='SetPageCounter 1 arabic'))
-        self.document.insert(0,nodes.raw(text='OddPageBreak cutePage'))
+        self.document.insert(0,nodes.raw(text='SetPageCounter 1 arabic', format='pdf'))
+        self.document.insert(0,nodes.raw(text='OddPageBreak %s'%self.page_template, format='pdf'))
         self.document.insert(0,contents)
-        self.document.insert(0,nodes.raw(text='SetPageCounter 1 lowerroman'))
+        self.document.insert(0,nodes.raw(text='SetPageCounter 1 lowerroman', format='pdf'))
         contTrans=PDFContents(self.document)
         contTrans.startnode=pending
         contTrans.apply()
@@ -533,7 +549,7 @@ class PDFWriter(writers.Writer):
             date=nodes.paragraph()
             date.append(nodes.Text(ustrftime(self.config.today_fmt or _('%B %d, %Y'))))
             date['classes']=['author']
-            self.document.insert(0,nodes.raw(text='OddPageBreak cutePage'))
+            self.document.insert(0,nodes.raw(text='OddPageBreak %s'%self.page_template, format='pdf'))
             self.document.insert(0,date)
             self.document.insert(0,spacer)
             for node in authornodes[::-1]:
@@ -545,6 +561,11 @@ class PDFWriter(writers.Writer):
         
         
         sio=StringIO()
+        
+        if self.invariant:
+            createpdf.patch_PDFDate()
+            createpdf.patch_digester()
+        
         createpdf.RstToPdf(sphinx=True,
                  stylesheets=self.stylesheets,
                  language=self.__language,
@@ -574,7 +595,8 @@ class PDFTranslator(nodes.SparseNodeVisitor):
         self.builder = builder
         self.footnotestack = []
         self.curfilestack = []
-	self.top_sectionlevel = 1
+        self.highlightlinenothreshold = 999999
+        self.top_sectionlevel = 1
         self.footnotecounter=1
         self.curfile=None
         self.footnotedict={}
@@ -621,13 +643,22 @@ class PDFTranslator(nodes.SparseNodeVisitor):
             pass
         else:
             lang=lang_for_block(node.astext(),node.get('language',self.highlightlang))
+            content = node.astext().splitlines()
+            if len(content) > self.highlightlinenothreshold or\
+               node.get('linenos',False):
+                options = { 'linenos': True }
+            else:
+                options = {}
+                
+            # FIXME: make tab width configurable
+            content = [c.replace('\t','        ') for c in content]
             replacement = nodes.literal_block()
             replacement.children = \
                 pygments_code_block_directive.code_block_directive(
                                     name = None,
                                     arguments = [lang],
-                                    options = {},
-                                    content = node.astext().splitlines(),
+                                    options = options,
+                                    content = content,
                                     lineno = False,
                                     content_offset = None,
                                     block_text = None,
@@ -656,6 +687,45 @@ class PDFTranslator(nodes.SparseNodeVisitor):
     def depart_desc_annotation(self, node):
         pass
     
+    # This is for graphviz support
+    def visit_graphviz(self, node):
+        # Not neat, but I need to send self to my handlers
+        node['builder']=self
+
+    def visit_Aanode(self, node):
+        pass
+    
+    def depart_Aanode(self, node):
+        pass
+
+    def visit_productionlist(self, node):
+        replacement=nodes.literal_block(classes=["code"])
+        names = []
+        for production in node:
+            names.append(production['tokenname'])
+        maxlen = max(len(name) for name in names)
+        for production in node:
+            if production['tokenname']:
+                lastname = production['tokenname'].ljust(maxlen)
+                n=nodes.strong()
+                n+=nodes.Text(lastname)
+                replacement+=n
+                replacement+=nodes.Text(' ::= ')
+            else:
+                replacement+=nodes.Text('%s     ' % (' '*len(lastname)))
+            production.walkabout(self)
+            replacement.children.extend(production.children)
+            replacement+=nodes.Text('\n')
+        node.parent.replace(node,replacement)
+        raise nodes.SkipNode
+    def depart_productionlist(self, node):
+        pass
+
+    def visit_production(self, node):
+        pass
+    def depart_production(self, node):
+        pass
+
 # This is copied from sphinx.highlighting
 def lang_for_block(source,lang):
     if lang in ('py', 'python'):
@@ -713,8 +783,46 @@ def try_parse(src):
     else:
         return True
 
+def init_math(app):
+    """
+        This is a dummy math extension.
+        
+        It's a hack, but if you want math in a PDF via pdfbuilder, and don't want to
+        enable pngmath or jsmath, then enable this one.
+        
+        :copyright: Copyright 2007-2009 by the Sphinx team, see AUTHORS.
+        :license: BSD, see LICENSE for details.
+    """
+    from sphinx.errors import SphinxError
+    try:
+        # Sphinx 0.6.4 and later
+        from sphinx.ext.mathbase import setup_math as mathbase_setup
+    except ImportError:
+        try:
+            # Sphinx 0.6.3
+            from sphinx.ext.mathbase import setup as mathbase_setup
+        except ImportError, e:
+            rst2pdf.log.log.error('Error importing sphinx math extension: %s', e)
+        
+    class MathExtError(SphinxError):
+        category = 'Math extension error'
+
+
+    def html_visit_math(self, node):
+        self.body.append(node['latex'])
+        raise nodes.SkipNode
+
+    def html_visit_displaymath(self, node):
+        self.body.append(node['latex'])
+        raise nodes.SkipNode
+
+    mathbase_setup(app, (html_visit_math, None), (html_visit_displaymath, None))
+
 
 def setup(app):
+    #Init dummy math extension
+    init_math(app)
+    
     app.add_builder(PDFBuilder)
     # PDF options
     app.add_config_value('pdf_documents', [], None)
@@ -733,6 +841,9 @@ def setup(app):
     app.add_config_value('pdf_splittables', False, None)
     app.add_config_value('pdf_breakside', 'odd', None)
     app.add_config_value('pdf_default_dpi', 300, None)
+    app.add_config_value('pdf_extensions',[], None)
+    app.add_config_value('pdf_page_template','cutePage', None)
+    app.add_config_value('pdf_invariant','False', None)
     
     author_texescaped = unicode(app.config.copyright)\
                                .translate(texescape.tex_escape_map)
