@@ -1,8 +1,8 @@
 # -*- coding: utf-8 -*-
 # See LICENSE.txt for licensing terms
-#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.12/rst2pdf/styles.py $
-#$LastChangedDate: 2009-10-21 09:20:04 -0300 (Wed, 21 Oct 2009) $
-#$LastChangedRevision: 1236 $
+#$HeadURL: https://rst2pdf.googlecode.com/svn/branches/0.14/rst2pdf/styles.py $
+#$LastChangedDate: 2010-03-20 12:04:22 -0300 (Sat, 20 Mar 2010) $
+#$LastChangedRevision: 1967 $
 
 import os
 import sys
@@ -24,21 +24,14 @@ from reportlab.lib.enums import *
 from reportlab.pdfbase import pdfmetrics
 import reportlab.lib.pagesizes as pagesizes
 
-try:
-    from json import loads
-except ImportError:
-    from simplejson import loads
+from rst2pdf.rson import loads as rson_loads
 
 import findfonts
 from log import log
 
-try:
-    from wordaxe.rl.paragraph import Paragraph
-    from wordaxe.rl.styles import ParagraphStyle, getSampleStyleSheet
-    from wordaxe import version as wordaxe_version
-    HAS_WORDAXE=True
-except ImportError:
-    HAS_WORDAXE=False
+from opt_imports import ParagraphStyle, wordaxe, wordaxe_version
+
+HAS_WORDAXE = wordaxe is not None
 
 unit_separator = re.compile('(-?[0-9\.]*)')
 
@@ -46,6 +39,44 @@ unit_separator = re.compile('(-?[0-9\.]*)')
 class StyleSheet(object):
     '''Class to handle a collection of stylesheets'''
 
+    @staticmethod
+    def stylepairs(data):
+        ''' Allows pairs of style information to be expressed
+            in canonical reportlab list of two-item list/tuple,
+            or in a more human-readable dictionary.
+        '''
+        styles = data.get('styles', {})
+        try:
+            stylenames = styles.keys()
+        except AttributeError:
+            for style in styles:
+                yield style
+            return
+
+        # Traditional reportlab styles are in ordered (key, value)
+        # tuples.  We also support dictionary lookup.  This is not
+        # necessarily ordered.
+
+        # The only problem with dictionary lookup is that
+        # we need to insure that parents are processed before
+        # their children.  This loop is a little ugly, but
+        # gets the job done.
+
+        while stylenames:
+            name = stylenames.pop()
+            parent = styles[name].get('parent')
+            if parent not in stylenames:
+                yield name, styles[name]
+                continue
+            names = [name]
+            while parent in stylenames:
+                stylenames.remove(parent)
+                names.append(parent)
+                parent = styles[names[-1]].get('parent')
+            while names:
+                name = names.pop()
+                yield name, styles[name]
+
     def __init__(self, flist, font_path=None, style_path=None, def_dpi=300):
         log.info('Using stylesheets: %s' % ','.join(flist))
         # find base path
@@ -53,28 +84,25 @@ class StyleSheet(object):
             self.PATH = abspath(dirname(sys.executable))
         else:
             self.PATH = abspath(dirname(__file__))
-            
+
         # flist is a list of stylesheet filenames.
         # They will be loaded and merged in order.
         # but the two default stylesheets will always
         # be loaded first
-        flist = [join(self.PATH, 'styles', 'styles.json'),
-                join(self.PATH, 'styles', 'default.json')] + flist
+        flist = [join(self.PATH, 'styles', 'styles.style'),
+                join(self.PATH, 'styles', 'default.style')] + flist
 
-        self.def_dpi=def_dpi        
-        dirn=os.path.dirname(__file__)
+        self.def_dpi=def_dpi
         if font_path is None:
             font_path=[]
-        font_path+=['.', os.path.join(os.path.abspath(dirn), 'fonts')]
+        font_path+=['.', os.path.join(self.PATH, 'fonts')]
         self.FontSearchPath = map(os.path.expanduser, font_path)
 
         if style_path is None:
             style_path=[]
-        style_path+=['.', os.path.join(os.path.abspath(dirn), 'styles'),
+        style_path+=['.', os.path.join(self.PATH, 'styles'),
                       '~/.rst2pdf/styles']
         self.StyleSearchPath = map(os.path.expanduser, style_path)
-
-
         self.FontSearchPath=list(set(self.FontSearchPath))
         self.StyleSearchPath=list(set(self.StyleSearchPath))
 
@@ -103,54 +131,13 @@ class StyleSheet(object):
         self.emsize=10
 
         self.languages = []
-        ssdata = []
-        # First parse all files and store the results
-        for fname in flist:
-            fname = self.findStyle(fname)
-            try:
-                if fname:
-                    ssdata.append(loads(open(fname).read()))
-            except ValueError, e: # Error parsing the JSON data
-                log.critical('Error parsing stylesheet "%s": %s'%\
-                    (fname, str(e)))
-                continue
-            except IOError, e: #Error opening the ssheet
-                log.critical('Error opening stylesheet "%s": %s'%\
-                    (fname, str(e)))
-                continue
 
-        # If any stylesheet has a [options][stylesheets] then those need to
-        # be inserted, too
-        _flist = []
-        for data, ssname in zip(ssdata, flist):
-            if 'options' in data and 'stylesheets' in data['options']:
-                _flist.extend(data['options']['stylesheets'])
-            _flist.append(ssname)
-        
-        if _flist != flist: # Need to reparse
-            flist = _flist
-            ssdata = []
-            # First parse all files and store the results
-            for fname in flist:
-                fname = self.findStyle(fname)
-                try:
-                    if fname:
-                        ssdata.append(loads(open(fname).read()))
-                except ValueError, e: # Error parsing the JSON data
-                    log.critical('Error parsing stylesheet "%s": %s'%\
-                        (fname, str(e)))
-                    continue
-                except IOError, e: #Error opening the ssheet
-                    log.critical('Error opening stylesheet "%s": %s'%\
-                        (fname, str(e)))
-                    continue
-        
-            
-            
+        ssdata = self.readSheets(flist)
+
         # Get pageSetup data from all stylessheets in order:
         self.ps = pagesizes.A4
         self.page={}
-        for data, ssname in zip(ssdata, flist):
+        for data, ssname in ssdata:
             page = data.get('pageSetup', {})
             if page:
                 self.page.update(page)
@@ -201,7 +188,7 @@ class StyleSheet(object):
 
         # Get page templates from all stylesheets
         self.pageTemplates = {}
-        for data, ssname in zip(ssdata, flist):
+        for data, ssname in ssdata:
             templates = data.get('pageTemplates', {})
             # templates is a dictionary of pageTemplates
             for key in templates:
@@ -215,13 +202,13 @@ class StyleSheet(object):
 
         # Get font aliases from all stylesheets in order
         self.fontsAlias = {}
-        for data, ssname in zip(ssdata, flist):
+        for data, ssname in ssdata:
             self.fontsAlias.update(data.get('fontsAlias', {}))
 
         embedded_fontnames = []
         self.embedded = []
         # Embed all fonts indicated in all stylesheets
-        for data, ssname in zip(ssdata, flist):
+        for data, ssname in ssdata:
             embedded = data.get('embeddedFonts', [])
 
             for font in embedded:
@@ -322,10 +309,8 @@ class StyleSheet(object):
 
         # Go though all styles in all stylesheets and find all fontNames.
         # Then decide what to do with them
-        for data, ssname in zip(ssdata, flist):
-            styles = data.get('styles', {})
-            for [skey, style] in styles:
-
+        for data, ssname in ssdata:
+            for [skey, style] in self.stylepairs(data):
                 for key in style:
                     if key == 'fontName' or key.endswith('FontName'):
                         # It's an alias, replace it
@@ -388,15 +373,20 @@ class StyleSheet(object):
                                       "replacing with Helvetica", style[key])
                             style[key] = "Helvetica"
         #log.info('FontList: %s'%self.embedded)
-        #log.info('FontAlias: %s'%self.fontsAlias)        
+        #log.info('FontAlias: %s'%self.fontsAlias)
         # Get styles from all stylesheets in order
         self.stylesheet = {}
         self.styles = []
         self.linkColor = 'navy'
-        for data, ssname in zip(ssdata, flist):
+        # FIXME: linkColor should probably not be a global
+        #        style, and tocColor should probably not
+        #        be a special case, but for now I'm going
+        #        with the flow...
+        self.tocColor = None
+        for data, ssname in ssdata:
             self.linkColor = data.get('linkColor') or self.linkColor
-            styles = data.get('styles', {})
-            for [skey, style] in styles:
+            self.tocColor = data.get('tocColor') or self.tocColor
+            for [skey, style] in self.stylepairs(data):
                 sdict = {}
                 # FIXME: this is done completely backwards
                 for key in style:
@@ -414,16 +404,25 @@ class StyleSheet(object):
                                 #command[3]=[str(c) for c in command[3]]
                             #elif c in ['BOX','INNERGRID'] or c.startswith('LINE'):
                                 #command[4]=str(command[4])
-                                    
+
                     # Handle alignment constants
                     elif key == 'alignment':
-                        style[key] = dict(TA_LEFT=0, TA_CENTER=1, TA_CENTRE=1,
-                            TA_RIGHT=2, TA_JUSTIFY=4, DECIMAL=8)[style[key]]
+                        style[key] = dict(TA_LEFT=0,
+                                          LEFT=0,
+                                          TA_CENTER=1,
+                                          CENTER=1,
+                                          TA_CENTRE=1,
+                                          CENTRE=1,
+                                          TA_RIGHT=2,
+                                          RIGHT=2,
+                                          TA_JUSTIFY=4,
+                                          JUSTIFY=4,
+                                          DECIMAL=8, )[style[key].upper()]
 
                     elif key == 'language':
                         if not style[key] in self.languages:
                             self.languages.append(style[key])
-                            
+
                     # Make keys str instead of unicode (required by reportlab)
                     sdict[str(key)] = style[key]
                     sdict['name'] = skey
@@ -481,7 +480,7 @@ class StyleSheet(object):
             # If the borderPadding is a list and wordaxe <=0.3.2,
             # convert it to an integer. Workaround for Issue
             if 'borderPadding' in s and ((HAS_WORDAXE and \
-                    wordaxe_version <='wordaxe 0.3.2') or 
+                    wordaxe_version <='wordaxe 0.3.2') or
                     reportlab.Version < "2.3" )\
                     and isinstance(s['borderPadding'], list):
                 log.warning('Using a borderPadding list in '\
@@ -499,7 +498,7 @@ class StyleSheet(object):
             if key.startswith('pygments'):
                 log.info("Using undefined style '%s'"
                             ", aliased to style 'code'."%key)
-                newst=copy(self.StyleSheet['code'])            
+                newst=copy(self.StyleSheet['code'])
             else:
                 log.warning("Using undefined style '%s'"
                             ", aliased to style 'normal'."%key)
@@ -508,6 +507,53 @@ class StyleSheet(object):
             self.StyleSheet.add(newst)
             return newst
 
+    def readSheets(self, flist):
+        ''' Read in the stylesheets.  Return a list of
+            (sheetdata, sheetname) tuples.
+
+            Orders included sheets in front
+            of including sheets.
+        '''
+        # Process from end of flist
+        flist.reverse()
+        # Keep previously seen sheets in sheetdict
+        sheetdict = {}
+        result = []
+
+        while flist:
+            ssname = flist.pop()
+            data = sheetdict.get(ssname)
+            if data is None:
+                data = self.readStyle(ssname)
+                if data is None:
+                    continue
+                sheetdict[ssname] = data
+                if 'options' in data and 'stylesheets' in data['options']:
+                    flist.append(ssname)
+                    newsheets = list(data['options']['stylesheets'])
+                    newsheets.reverse()
+                    flist.extend(newsheets)
+                    continue
+            result.append((data, ssname))
+        return result
+
+    def readStyle(self, ssname):
+            # If callables are used, they should probably be subclassed
+            # strings, or something else that will print nicely for errors
+            if callable(ssname):
+                return ssname()
+
+            fname = self.findStyle(ssname)
+            if fname:
+                try:
+                    return rson_loads(open(fname).read())
+                except ValueError, e: # Error parsing the JSON data
+                    log.critical('Error parsing stylesheet "%s": %s'%\
+                        (fname, str(e)))
+                except IOError, e: #Error opening the ssheet
+                    log.critical('Error opening stylesheet "%s": %s'%\
+                        (fname, str(e)))
+
     def findStyle(self, fn):
         """Find the absolute file name for a given style filename.
 
@@ -673,6 +719,29 @@ class StyleSheet(object):
                            default_unit,
                            emsize=self.emsize)
 
+    def combinedStyle(self, styles):
+        '''Given a list of style names, it merges them (the existing ones)
+        and returns a new style.
+
+        The styles that don't exist are silently ignored.
+
+        For example, if called with styles=['style1','style2'] the returned
+        style will be called 'merged_style1_style2'.
+
+        The styles that are *later* in the list will have priority.
+        '''
+
+        validst = [x for x in styles if self.StyleSheet.has_key(x)]
+        newname = '_'.join(['merged']+validst)
+        validst = [self[x] for x in validst]
+        newst=copy(validst[0])
+
+        for st in validst[1:]:
+            newst.__dict__.update(st.__dict__)
+        
+        newst.name=newname
+        return newst
+        
 
 def adjustUnits(v, total=None, dpi=300, default_unit='pt', emsize=10):
     """Takes something like 2cm and returns 2*cm.
@@ -734,6 +803,9 @@ def formatColor(value, numeric=True):
             r = int(c[:2], 16)/255.
             g = int(c[2:4], 16)/255.
             b = int(c[4:6], 16)/255.
+            if len(c) >= 8:
+                alpha = int(c[6:8], 16)/255.
+                return colors.Color(r, g, b, alpha=alpha)
             return colors.Color(r, g, b)
         else:
             return str("#"+c)
@@ -743,7 +815,7 @@ def formatColor(value, numeric=True):
 # * Maximum number of arguments
 # * Valid types of arguments.
 #
-# For example, if option FOO takes a list a string and a number, 
+# For example, if option FOO takes a list a string and a number,
 # but the number is optional:
 #
 # "FOO":(2,3,"list","string","number")
@@ -752,7 +824,7 @@ def formatColor(value, numeric=True):
 #
 # ["FOO",(0,0),(-1,-1),[1,2],"whatever",4]
 #
-# THe (0,0) (-1,-1) are start and stop and are mandatory. 
+# THe (0,0) (-1,-1) are start and stop and are mandatory.
 #
 # Possible types of arguments are string, number, color, colorlist
 
@@ -792,9 +864,9 @@ validCommands={
 def validateCommands(commands):
     '''Given a list of reportlab's table commands, it fixes some common errors
     and/or removes commands that can't be fixed'''
-    
+
     fixed=[]
-    
+
     for command in commands:
         command[0]=command[0].upper()
         flag=False
@@ -802,26 +874,26 @@ def validateCommands(commands):
         if command[0] not in validCommands:
             log.error('Unknown table command %s in stylesheet',command[0])
             continue
-        
+
         # See if start and stop are the right types
         if type(command[1]) not in (ListType,TupleType):
             log.error('Start cell in table command should be list or tuple, got %s [%s]',type(command[1]),command[1])
             flag=True
-            
+
         if type(command[2]) not in (ListType,TupleType):
             log.error('Stop cell in table command should be list or tuple, got %s [%s]',type(command[1]),command[1])
             flag=True
-        
+
         # See if the number of arguments is right
         l=len(command)-3
         if l>validCommands[command[0]][1]:
             log.error('Too many arguments in table command: %s',command)
             flag=True
-            
+
         if l<validCommands[command[0]][0]:
             log.error('Too few arguments in table command: %s',command)
             flag=True
-            
+
         # Validate argument types
         for pos,arg in enumerate(command[3:]):
             typ = validCommands[command[0]][pos+2]
@@ -837,8 +909,21 @@ def validateCommands(commands):
                 command[3+pos]=str(arg)
             else:
                 log.error("This should never happen: wrong type %s",typ)
-            
+
         if not flag:
             fixed.append(command)
-            
+
     return fixed
+
+class CallableStyleSheet(str):
+    ''' Useful for programmatically generated stylesheets.
+        A generated stylesheet is a callable string (name),
+        which returns the pre-digested stylesheet data
+        when called.
+    '''
+    def __new__(cls, name, value=''):
+        self = str.__new__(cls, name)
+        self.value = value
+        return self
+    def __call__(self):
+        return rson_loads(self.value)
